@@ -1,10 +1,11 @@
 from .figure import Figure
+
 try:
-	from importlib.metadata import version
+    from importlib.metadata import version
 except ModuleNotFoundError:
     from importlib_metadata import version
 
 try:
     __version__ = version(__name__)
-except:
+except NameError:
     pass
