@@ -1,15 +1,17 @@
-from .scales import *
-from . import utils
-from .img2ascii import img2ascii
-from .braille import is_braille, draw_braille
-from warnings import warn
-from typing import Optional
+from functools import partial
+from numbers import Number
 from shutil import get_terminal_size
+from typing import Callable, Iterable, List, Optional, Tuple
+
 import numpy as np
-from functools import lru_cache, partial
-from termcolor import colored
+from backports.cached_property import cached_property
 from colorama import init
-from copy import deepcopy
+from termcolor import colored
+
+from . import utils
+from .braille import draw_braille, is_braille
+from .img2ascii import img2ascii
+from .scales import CategoricalScale, LinearScale
 
 init()
 
@@ -41,7 +43,7 @@ class Figure:
         height: Height of the figure in number of characters. Defaults to the terminal window height, or falls back to 24.
         legendloc: Legend location. Supported values are `"topleft"`, `"topright"`, `"bottomleft"`, and `"bottomright"`.
         ascii: Set to `True` to only use ascii characters. Defaults to trying to detect if unicode is supported in the terminal.
-        y_axis_direction: Set to `"up"` to have Y axis point up (conventional for graphs), `"down"` to have Y axis point down 
+        y_axis_direction: Set to `"up"` to have Y axis point up (conventional for graphs), `"down"` to have Y axis point down
                           (conventional for images). By default, this is automatically determined based on the drawn plots.
     """
 
@@ -55,7 +57,7 @@ class Figure:
         legendloc: str = "topright",
         ascii: bool = False,
         y_axis_direction: str = "auto",
-    ):
+    ) -> None:
         if legendloc not in {"topleft", "topright", "bottomleft", "bottomright"}:
             raise ValueError("Unsupported legend location")
         if width is not None:
@@ -80,8 +82,8 @@ class Figure:
         self.height = height if height else term_height
 
         # gather stuff to plot before actually drawing it
-        self._plots = []
-        self._labels = []
+        self._plots: List[Callable] = []
+        self._labels: List[Tuple[str, str]] = []
 
     @property
     def _x(self):
@@ -91,7 +93,7 @@ class Figure:
     def _y(self):
         return tuple([y for plot in self._plots for y in plot.keywords["y"]])
 
-    @lru_cache(maxsize=1)
+    @cached_property
     def _yscale(self):
         if utils._is_numerical(self._y):
             scale = LinearScale()
@@ -104,39 +106,39 @@ class Figure:
         scale.fit(self._y, target_min, target_max)
         if utils._is_numerical(self._y):
             # refit scale to tick values, since those lay just outside the input data range
-            scale.fit(self._ytick_values(), target_min, target_max)
+            scale.fit(self._ytick_values, target_min, target_max)
         return scale
 
-    @lru_cache(maxsize=1)
+    @cached_property
     def _xscale(self):
         if utils._is_numerical(self._x):
             scale = LinearScale()
         else:
             scale = CategoricalScale()
-        target_min = self._yax_width()
+        target_min = self._yax_width
         target_max = self.width - 1
         scale.fit(self._x, target_min, target_max)
         if utils._is_numerical(self._x):
             # refit scale to tick values, since those lay just outside the input data range
-            scale.fit(self._xtick_values(), target_min, target_max)
+            scale.fit(self._xtick_values, target_min, target_max)
         return scale
 
-    def _xax_height(self):
+    def _xax_height(self) -> int:
         return 2 + bool(self._xlabel)
 
-    def _fmt(self, value):
+    def _fmt(self, value) -> str:
         if isinstance(value, Number):
             return f"{value:.3g}"
         else:
             return str(value)
 
-    @lru_cache(maxsize=1)
-    def _yax_width(self):
+    @cached_property
+    def _yax_width(self) -> int:
         """
         Since y-axis tick labels are drawn horizontally, the width of the y axis
         depends on the length of the labels, which themselves depend on the data.
         """
-        labels = (self._fmt(value) for value in self._ytick_values())
+        labels = (self._fmt(value) for value in self._ytick_values)
         width = max([len(label) for label in labels])
         width += 1  # for axis ticks
         width += bool(self._ylabel) * 2  # for y label
@@ -151,12 +153,12 @@ class Figure:
     def _rjust_draw(self, string, array, fillchar=" "):
         array[:] = list(string.rjust(len(array), fillchar))
 
-    @lru_cache(maxsize=1)
+    @cached_property
     def _ytick_values(self):
         if utils._is_numerical(self._y):
             return utils._best_ticks(min(self._y), max(self._y), most=self.height // 3)
         else:  # nominal
-            values = sorted([str(v) for v in set(self._y)])
+            values = tuple(sorted([str(v) for v in set(self._y)]))
             y_axis_height = self.height - bool(self.title) - self._xax_height()
             if len(values) > y_axis_height:
                 raise ValueError(
@@ -164,42 +166,40 @@ class Figure:
                 )
             return values
 
-    @lru_cache(maxsize=1)
+    @cached_property
     def _xtick_values(self):
         if utils._is_numerical(self._x):
             return utils._best_ticks(min(self._x), max(self._x), most=self.width // 4)
         else:  # categorical
             # note this may not fit depending on the width of the figure
-            values = sorted([str(v) for v in set(self._x)])
+            values = tuple(sorted([str(v) for v in set(self._x)]))
             return values
 
-    def _draw_y_axis(self):
-        start = round(self._yscale().transform(self._ytick_values()[-1]))
-        end = round(self._yscale().transform(self._ytick_values()[0]))
+    def _draw_y_axis(self) -> None:
+        start = round(self._yscale.transform(self._ytick_values[-1]))
+        end = round(self._yscale.transform(self._ytick_values[0]))
         start, end = min(start, end), max(start, end)
-        self._canvas[start:end, self._yax_width() - 1] = "│"
+        self._canvas[start:end, self._yax_width - 1] = "│"
         for value, pos in zip(
-            self._ytick_values(), self._yscale().transform(self._ytick_values())
+            self._ytick_values, self._yscale.transform(self._ytick_values)
         ):
             pos = round(pos)
             label = self._fmt(value)
-            self._canvas[pos, self._yax_width() - 1] = "┤"
+            self._canvas[pos, self._yax_width - 1] = "┤"
             self._rjust_draw(
-                label, self._canvas[pos, bool(self._ylabel) * 2 : self._yax_width() - 1]
+                label, self._canvas[pos, bool(self._ylabel) * 2 : self._yax_width - 1]
             )
 
         if self._ylabel:
             ylabel = self._ylabel[: end - start]  # make sure it fits
             self._center_draw(ylabel, self._canvas[start:end, 0])
 
-    def _draw_x_axis(self):
-        tick_positions = [
-            round(v) for v in self._xscale().transform(self._xtick_values())
-        ]
-        labels = [self._fmt(v) for v in self._xtick_values()]
+    def _draw_x_axis(self) -> None:
+        tick_positions = [round(v) for v in self._xscale.transform(self._xtick_values)]
+        labels = [self._fmt(v) for v in self._xtick_values]
         # draw axis
-        axis_start = round(self._xscale().transform(self._xtick_values()[0]))
-        axis_end = round(self._xscale().transform(self._xtick_values()[-1]))
+        axis_start = round(self._xscale.transform(self._xtick_values[0]))
+        axis_end = round(self._xscale.transform(self._xtick_values[-1]))
         self._canvas[-self._xax_height(), axis_start:axis_end] = "─"
         # draw ticks
         for tick_pos in tick_positions:
@@ -216,20 +216,19 @@ class Figure:
             xlabel = self._xlabel[: axis_end - axis_start]  # make sure it fits
             self._center_draw(xlabel, self._canvas[-1, axis_start:axis_end])
 
-    def _draw_legend(self):
-        # labelstrings = [f"{marker} {label}" for marker, label in self._labels]
+    def _draw_legend(self) -> None:
         width = max([len(label) for marker, label in self._labels]) + 4
         width = max(width, len("Legend") + 2)
         height = len(self._labels) + 2
 
         if self.legendloc.startswith("top"):
-            top = int(self._yscale().transform(self._ytick_values()[-1]))
+            top = int(self._yscale.transform(self._ytick_values[-1]))
         elif self.legendloc.startswith("bottom"):
-            top = int(self._yscale().transform(self._ytick_values()[0])) - height + 1
+            top = int(self._yscale.transform(self._ytick_values[0])) - height + 1
         if self.legendloc.endswith("right"):
-            left = int(self._xscale().transform(self._xtick_values()[-1])) - width + 1
+            left = int(self._xscale.transform(self._xtick_values[-1])) - width + 1
         elif self.legendloc.endswith("left"):
-            left = int(self._xscale().transform(self._xtick_values()[0]))
+            left = int(self._xscale.transform(self._xtick_values[0]))
 
         self._canvas[top, left : left + width] = list(
             "┌" + "Legend".center(width - 2, "─") + "┐"
@@ -244,8 +243,8 @@ class Figure:
             "└" + "─" * (width - 2) + "┘"
         )
 
-    def _prep(self, x, y, marker, color, label):
-        """ Data preparation stuff common to all plots. """
+    def _prep(self, x, y, marker, color, label) -> tuple:
+        """Data preparation stuff common to all plots."""
         x_is_valid = x is not None and len(x) > 0
         y_is_valid = y is not None and len(y) > 0
         if not x_is_valid and not y_is_valid:
@@ -279,7 +278,7 @@ class Figure:
         marker: str = "•",
         color: Optional[str] = None,
         label: Optional[str] = None,
-    ):
+    ) -> None:
         """
         Adds scatter plot.
 
@@ -293,7 +292,7 @@ class Figure:
         x, y, marker, color, label = self._prep(x, y, marker, color, label)
 
         def draw_scatter(x, y, marker):
-            for xi, yi in zip(self._xscale().transform(x), self._yscale().transform(y)):
+            for xi, yi in zip(self._xscale.transform(x), self._yscale.transform(y)):
                 if not self.ascii_only and any((is_braille(char) for char in marker)):
                     xi = utils._round_half_away_from_zero(xi)
                     yi = utils._round_half_away_from_zero(yi)
@@ -313,7 +312,7 @@ class Figure:
         marker: str = "braille",
         color: Optional[str] = None,
         label: Optional[str] = None,
-    ):
+    ) -> None:
         """
         Adds line plot.
 
@@ -327,8 +326,8 @@ class Figure:
         x, y, marker, color, label = self._prep(x, y, marker, color, label)
 
         def draw_line(x, y, marker):
-            xs = self._xscale().transform(x)
-            ys = self._yscale().transform(y)
+            xs = self._xscale.transform(x)
+            ys = self._yscale.transform(y)
             for (x0, x1), (y0, y1) in zip(zip(xs[:-1], xs[1:]), zip(ys[:-1], ys[1:])):
                 if not self.ascii_only and any((is_braille(char) for char in marker)):
                     for x, y in utils._plot_line_segment(
@@ -357,7 +356,7 @@ class Figure:
         marker: str = "█",
         color: Optional[str] = None,
         label: Optional[str] = None,
-    ):
+    ) -> None:
         """
         Adds vertical bar plot.
 
@@ -373,10 +372,10 @@ class Figure:
         def draw_bar(x, y, marker):
             marker = marker.replace("⠄", "⡇")  # in case of braille
             if utils._is_numerical(self._y):
-                origin = self._yscale().transform(min(self._ytick_values(), key=abs))
+                origin = self._yscale.transform(min(self._ytick_values, key=abs))
             else:
-                origin = self._yscale().transform(self._ytick_values()[0])
-            for xi, yi in zip(self._xscale().transform(x), self._yscale().transform(y)):
+                origin = self._yscale.transform(self._ytick_values[0])
+            for xi, yi in zip(self._xscale.transform(x), self._yscale.transform(y)):
                 start, end = sorted([origin, yi])
                 self._canvas[round(start) : round(end) + 1, round(xi)] = marker
 
@@ -389,7 +388,7 @@ class Figure:
         marker: str = "█",
         color: Optional[str] = None,
         label: Optional[str] = None,
-    ):
+    ) -> None:
         """
         Adds horizontal bar plot.
 
@@ -405,16 +404,16 @@ class Figure:
         def draw_hbar(x, y, marker):
             marker = marker.replace("⠄", "⠒")  # in case of braille
             if utils._is_numerical(self._x):
-                origin = self._xscale().transform(min(self._xtick_values(), key=abs))
+                origin = self._xscale.transform(min(self._xtick_values, key=abs))
             else:
-                origin = self._xscale().transform(self._xtick_values()[0])
-            for xi, yi in zip(self._xscale().transform(x), self._yscale().transform(y)):
+                origin = self._xscale.transform(self._xtick_values[0])
+            for xi, yi in zip(self._xscale.transform(x), self._yscale.transform(y)):
                 start, end = sorted([origin, xi])
                 self._canvas[round(yi), round(start) : round(end) + 1] = marker
 
         self._plots.append(partial(draw_hbar, x=x, y=y, marker=marker))
 
-    def text(self, x, y, text: str, color: Optional[str] = None):
+    def text(self, x, y, text: str, color: Optional[str] = None) -> None:
         """
         Adds text.
 
@@ -428,8 +427,8 @@ class Figure:
             text = colored(text, color)
 
         def draw_text(x, y, text):
-            x0 = round(self._xscale().transform(x[0]))
-            y0 = round(self._yscale().transform(y[0]))
+            x0 = round(self._xscale.transform(x[0]))
+            y0 = round(self._yscale.transform(y[0]))
             for i, char in enumerate(text):
                 if x0 + i >= self.width:
                     break
@@ -443,7 +442,7 @@ class Figure:
         vmin: Optional[float] = None,
         vmax: Optional[float] = None,
         cmap: str = "block",
-    ):
+    ) -> None:
         """
         Adds image.
 
@@ -451,7 +450,7 @@ class Figure:
 
         Args:
             image: 2D array.
-            vmin: Minimum value covered by the colormap. Lower values are clipped. 
+            vmin: Minimum value covered by the colormap. Lower values are clipped.
                   If set to `None`, uses 0 if the `dtype` of image is `numpy.uint8` (usual for pictures), `min(image)` otherwise.
             vmax: Maximum value covered by the colormap. Higher values are clipped.
                   If set to `None`, uses 255 if the `dtype` of image is `numpy.uint8` (usual for pictures), `max(image)` otherwise.
@@ -473,10 +472,10 @@ class Figure:
             self._y_axis_direction = "down"
 
         def draw_image(x, y):
-            xmin = round(self._xscale().transform(0))
-            ymin = round(self._yscale().transform(0))
-            xmax = round(self._xscale().transform(image.shape[1]))
-            ymax = round(self._yscale().transform(image.shape[0]))
+            xmin = round(self._xscale.transform(0))
+            ymin = round(self._yscale.transform(0))
+            xmax = round(self._xscale.transform(image.shape[1]))
+            ymax = round(self._yscale.transform(image.shape[0]))
             ymin, ymax = min(ymin, ymax), max(ymin, ymax)
             drawn = img2ascii(
                 image,
@@ -499,7 +498,7 @@ class Figure:
         )
         self._clear_scale_cache()
 
-    def _draw(self):
+    def _draw(self) -> None:
         if not self._plots:
             raise ValueError("No plots to draw.")
 
@@ -523,24 +522,25 @@ class Figure:
             for old, new in ASCII_FALLBACK.items():
                 self._canvas = np.char.replace(self._canvas, old, new)
 
-    def clear(self):
-        """ Clears previously added plots. """
+    def clear(self) -> None:
+        """Clears previously added plots."""
         self._plots = []
         self._labels = []
         self._clear_scale_cache()
 
-    def _clear_scale_cache(self):
-        self._xscale.cache_clear()
-        self._yscale.cache_clear()
-        self._xtick_values.cache_clear()
-        self._ytick_values.cache_clear()
-        self._yax_width.cache_clear()
+    def _clear_scale_cache(self) -> None:
+        # clear cached values if cached, otherwise do nothing
+        self.__dict__.pop("_xscale", None)
+        self.__dict__.pop("_yscale", None)
+        self.__dict__.pop("_xtick_values", None)
+        self.__dict__.pop("_ytick_values", None)
+        self.__dict__.pop("_yax_width", None)
 
-    def __str__(self):
+    def __str__(self) -> str:
         self._draw()
         return "\n".join(["".join(row) for row in self._canvas.tolist()])
 
-    def show(self):
+    def show(self) -> None:
         """
         Prints the figure.
 
